public class RealLine {
	// the line Ax + By + C = 0
	double A, B, C;
	
	public RealLine(RealPoint P, RealPoint Q) {
		A = -(Q.y - P.y);
		B = Q.x - P.x;
		C = -(A*P.x + B*P.y);
	}
	
	public boolean isDegenerate() {
		return(A == 0.0 && B == 0.0);
	}
	
	public static RealPoint interp(double t, RealPoint P, RealPoint Q) {
		double x = (1.0-t)*P.x + t*Q.x;
		double y = (1.0-t)*P.y + t*Q.y;
		return(new RealPoint(x, y));
	}
	
	public double eval(double x, double y) {
		return(A*x + B*y + C);
	}
	
	public double eval(RealPoint P) {
		return(A*P.x + B*P.y + C);
	}
	
	// returns the segment intersecting the canvas boundaries
	// line MUST not be degenerate
	public PlotPath segmentAcross(PlotCanvas canvas) {
		// look at all four corners of the canvas
		// if (isDegenerate()) return(null);
		RealPoint c0 = canvas.ctm.toRealPoint(0, 0);
		RealPoint c1 = canvas.ctm.toRealPoint(canvas.width, 0);
		RealPoint c2 = canvas.ctm.toRealPoint(canvas.width, canvas.height);
		RealPoint c3 = canvas.ctm.toRealPoint(0, canvas.height);
		double r0 = eval(c0);
		double r1 = eval(c1);
		double r2 = eval(c2);
		double r3 = eval(c3);
		RealPoint P[] = new RealPoint[2];
		int count = 0;
		if ((r0 < 0.0 && r1 >= 0.0) 
			|| (r0 >= 0.0 && r1 < 0.0)) {
			// on top
			if (r1 != r0) 
				P[count++] = interp(-r0/(r1 - r0), c0, c1);
		} 
		if ((r1 < 0.0 && r2 >= 0.0) 
			|| (r1 >= 0.0 && r2 < 0.0)) {
			// on right
			if (r2 != r1) 
				P[count++] = interp(-r1/(r2 - r1), c1, c2);
		}
		if ((r2 < 0.0 && r3 >= 0.0) 
			|| (r2 >= 0.0 && r3 < 0.0)) {
			// on bottom
			if (r3 != r2) 
				P[count++] = interp(-r2/(r3 - r2), c2, c3);
		}
		if ((r3 < 0.0 && r0 >= 0.0) 
			|| (r3 >= 0.0 && r0 < 0.0)) {
			// on left
			if (r3 != r0) 
				P[count++] = interp(-r3/(r0 - r3), c3, c0);
		}
		if (count == 0) {
			return(null);
		} else if (count == 1) {
			PlotPath ell = new PlotPath(canvas);
			ell.moveto(P[0].x, P[0].y);
			ell.lineto(P[0].x, P[0].y);
			return(ell);
		} else {
			PlotPath ell = new PlotPath(canvas);
			ell.moveto(P[0].x, P[0].y);
			ell.lineto(P[1].x, P[1].y);
			return(ell);
		}
	}
	
	// both lines must be non-degenerate
	public RealPoint intersectionWith(RealLine ell) {
		double x;
		double y;
		double det = A*ell.B - ell.A*B;
		if (det == 0.0) return(null);
		y = -(A*ell.C - ell.A*C)/det;
		x =  (B*ell.C - ell.B*C)/det;
		return(new RealPoint(x, y));
	}
	
	// MUST be nondegenerate
	public RealPoint projectionFrom(RealPoint P) {
		if (isDegenerate()) return(null);
		double t = (A*P.x + B*P.y + C)/(A*A + B*B);
		double x = P.x - t*A;
		double y = P.y - t*B;
		RealPoint Q = new RealPoint(x, y);
		return(Q);
	}
}